def interface____differentiated__3() -> None:
    return None

mod = "django.contrib.contenttypes.models"
summary = "__virtual__.mod_3961720227::django.contrib.contenttypes.models::installed_apps=__installed_apps_hash__::significant=2877817555"

import django.contrib.contenttypes.models
import django.db.models
ConcreteQuerySet__ContentType = django.db.models.QuerySet[django.contrib.contenttypes.models.ContentType]
Concrete__ContentType = django.contrib.contenttypes.models.ContentType
