def interface____differentiated__6() -> None:
    return None

mod = "djangoexample.exampleapp2.models"
summary = "__virtual__.mod_3537308831::djangoexample.exampleapp2.models::installed_apps=__installed_apps_hash__::significant=1778215230"

import django.db.models
import djangoexample.exampleapp2.models
ConcreteQuerySet__ChildOther = django.db.models.QuerySet[djangoexample.exampleapp2.models.ChildOther]
ConcreteQuerySet__ChildOther2 = django.db.models.QuerySet[djangoexample.exampleapp2.models.ChildOther2]
Concrete__ChildOther = djangoexample.exampleapp2.models.ChildOther
Concrete__ChildOther2 = djangoexample.exampleapp2.models.ChildOther2
