def interface____differentiated__10() -> None:
    return None

mod = "djangoexample.empty_models.models"
summary = "__virtual__.mod_3808300370::djangoexample.empty_models.models::installed_apps=__installed_apps_hash__::significant=1232670738"
