def interface____differentiated__12() -> None:
    return None

mod = "django.contrib.sessions.base_session"
summary = "__virtual__.mod_113708644::django.contrib.sessions.base_session::installed_apps=__installed_apps_hash__::significant=1879365353"

import django.contrib.sessions.base_session
import django.contrib.sessions.models
import django.db.models
ConcreteQuerySet__AbstractBaseSession = django.db.models.QuerySet[django.contrib.sessions.models.Session]
Concrete__AbstractBaseSession = django.contrib.sessions.models.Session
