def interface____differentiated__4() -> None:
    return None

mod = "django.contrib.sessions.models"
summary = "__virtual__.mod_3074165738::django.contrib.sessions.models::installed_apps=__installed_apps_hash__::significant=4231118930"

import django.contrib.sessions.models
import django.db.models
ConcreteQuerySet__Session = django.db.models.QuerySet[django.contrib.sessions.models.Session]
Concrete__Session = django.contrib.sessions.models.Session
